/********************************************************
*	
*
*				by devoto
*
*   
********************************************************/
#include<math.h> //sign and cos stuff
#include"gba.h"
#include"screenmode.h"
#include"keypad.h"
#include"sprite.h"
#include"Backgrounds.h"

//my graphics data produced by pcx2sprite program
#include"ghosts.h"
#include"pacman.h"
#include"tiles.h"

//data produced by map editor
#include"level1.c"


///sprite structure

typedef struct tagSprite
{
    FIXED x,y;				//x and y coordinates
	char state;				//state (alive, dead, hurt ect...) 
	u16 oamEntry;			//oam entrie for the sprite graphics
	u16 frameCharName[8];	//char name of the start of each frame
	char curFrame;			//current frame
	u8 rot_index;			//index int RotData[]
}Sprite,*pSprite;

///declare our functions so the c compiler knows what we are talking about

void InitializeSprites(void);
void WaitForVsync(void);
void CopyOAM(void);
void GetInput(void);
void MoveSprite(pSprite sp);
void RotateSprite(int rotDataIndex, int angle, FIXED x_scale,FIXED y_scale);
void SetUpSprite(pSprite sp);
void MoveEnemies(pSprite sp);


///these are global so i do not have to pass them to GetInput

FIXED angle = 0;
FIXED zoom = 1<<8;  //zoom is a fixed point number

FIXED SIN[360];		//my LUTs for sign and cosign
FIXED COS[360];

char RotIndexCounter = 0;  //global to keep track of rotation indexes used

#define LVL_1_H 304   //height and width of my actual map data
#define LVL_1_W 432
//global sprites for easy access
Sprite pacman;
Sprite ghosts[4];
// some state defines so I know what my sprites are doing
#define ALIVE			0		//pacman alive and well
#define DEAD			1		//well dead i supose :)
#define EMPOWERED		2		//ate a power pellet
#define REGENERATING	3		//ghost is dead and being rebuilt in the pit
#define CHASEING		4		//ghost is chasing pacman
#define RUNNING			5		//ghost is running from pacman
#define DIEING			6		//pacman is dieing

//this is my background structure defined in backgrounds.h

Bg bg2;  

///main entry point from the boot.asm startup file
void AgbMain(void)
{
	int index = 0;  //some looping variables for loops :)
	u16 loop;
	u16* temp;
	
	//compute my Look up tables
	for(loop = 0; loop < 360; loop++)
	{
		SIN[loop] = (FIXED)(sin(RADIAN(loop)) * 256);  //sin and cos are computed and cast to fixed							//fixed
		COS[loop] = (FIXED)(cos(RADIAN(loop)) * 256);
	}

	//set mode 2 and enable sprites and 1d mapping
	SetMode(MODE_1 | OBJ_ENABLE | OBJ_MAP_1D);  

	//copy my palette in
	for(loop = 0; loop < 256; loop++)
		OBJPaletteMem[loop] = ghostsPalette[loop];	
									//OBJPaletteMem is defined in gba.h.
									//ghostsPalette is from ghosts.h
								
	
	InitializeSprites();  //set all 128 sprites to offscreen

//Let us set up the backgroud two structure and enable the background

	bg2.number = 2;				//background number 0-3
	bg2.charBaseBlock = 0;
	bg2.screenBaseBlock = 28;
	bg2.colorMode = BG_COLOR_256;
	bg2.size = ROTBG_SIZE_512x512;
	bg2.mosiac = 0;
	bg2.x_scroll = 120;
	bg2.y_scroll = 80;

//This call enables the background with the suplied attributes and when it returns
//bg2.mapdata and tiledata will be valid pointers to the correct place
//for the map and tile data in VRAM
	
	EnableBackground(&bg2);


//Now let us copy in the background tile palette
	for(loop = 0; loop < 256; loop++)
		BGPaletteMem[loop] =tilesPalette[loop];
//now for the actual tile graphics created by pcx2sprite 
	for(loop = 0; loop < tiles_WIDTH * tiles_HEIGHT /2; loop++)
		bg2.tileData[loop] =tilesData[loop];

//now for the map a temp variable is necessary because the map editor saves it as a
//u8 array.  The map data was padded to 64x64 tiles even though our map is not quite 
//that big.
	temp = (u16*)level1;  
	for(loop = 0; loop < 64*64/2; loop++) //64x64 tiles /2 because 16 bit copy
		bg2.mapData[loop] = temp[loop];

//first lets copy in pacmans bitmap data.  8 frames * 16x16 = 2048 0r 1024 double
//bytes

		for(loop = 0; loop < 1024; loop++)
		{
			OAMData[loop] = pacmanData[loop];  
			
		}//end loop

//now the ghost data.  4 frames * 16x16 = 512 .  We will leave loop at 1024 because 
//that is were pacmans data stoped.  We will just copy it on top.  If this were
//2d mode we would have to figure out were we wanted all our sprite date
//but 1d allows us to just keep stacking.

		for(loop = 1024; loop < (1024 + 512); loop++)
		{
			OAMData[loop] = ghostsData[loop-1024];  
			
		}//end loop


//Now we set up our sprites.  Since they are 16x16 256-color sprites they take
//8 tiles per sprite.  Remeber that a tile is 8x8 16-color sprite so it 
//takes 8.
	pacman.oamEntry = 0;			//pacman gets the first slot
	pacman.curFrame = 0;			//start facing right
	pacman.frameCharName[0] = 0;	//8 frames of animation U,D,L,R and open and closed mouth
	pacman.frameCharName[1] = 8;
	pacman.frameCharName[2] = 16;
	pacman.frameCharName[3] = 24;
	pacman.frameCharName[4] = 32;
	pacman.frameCharName[5] = 40;
	pacman.frameCharName[6] = 48;
	pacman.frameCharName[7] = 56;
	pacman.state = ALIVE;			//starts out alive
	pacman.x = 112<<8;				//x and y are fixed point although not really necessary
	pacman.y = 72<<8;
	pacman.rot_index = RotIndexCounter++;  //increment counter so the next sprite has a diferent one

//this function just copies our pacman structure to the corect place in our OAM copy
	SetUpSprite(&pacman);
	
//loop through all ghosts.  pick the next oam entry and the next char name.
	for(index = 0; index <4; index++)
	{
		ghosts[index].curFrame = 0;  //only one frame
		ghosts[index].frameCharName[0] = 64 + (8 * index); //next character name
		ghosts[index].oamEntry = index + 1; //next OAM
		ghosts[index].state = CHASEING; //state
		ghosts[index].x = (16 + 18 * index)<<8;  //Fixed point again
		ghosts[index].y = 16<<8;
		ghosts[index].rot_index = RotIndexCounter++; //same as above
		SetUpSprite(&ghosts[index]);
	}

///Main Game loop	
	while(1)
	{
	

		GetInput();    //get input changes the x and y based on input
		MoveSprite(&pacman); //changes sprite atributes based on new x,y
		RotateSprite(pacman.rot_index,angle,zoom,zoom);

		for(index = 0; index < 4; index++)
		{
			MoveEnemies(&ghosts[index]);//changes ghost x,y based on state
			MoveSprite(&ghosts[index]); //puts x,y into OAM entries
			RotateSprite(ghosts[index].rot_index,angle,zoom,zoom);
		}
		
		RotateBackground(&bg2,angle,119,79,zoom);
		WaitForVsync();			//waits for the screen to stop drawing
		CopyOAM();				//Copies our sprite array into OAM.
		UpdateBackground(&bg2); //makes any changes to the bg struct get copied inot the apropriate registers

	}
}

////Set sprites to off screen

void InitializeSprites(void)
{
	int loop;
	for(loop = 0; loop < 128; loop++)
	{
		sprites[loop].attribute0 = SQUARE | COLOR_256 | 160 | SIZE_DOUBLE;  //y to > 159
		sprites[loop].attribute1 = SIZE_16 | 240;  //x to > 239

	}
}

//wait for the screen to stop drawing
void WaitForVsync(void)
{
	/*__asm 
	{
		mov 	r0, #0x4000006   //0x4000006 is vertical trace counter; when it hits 160					 //160 the vblanc starts
		scanline_wait:	       	//the vertical blank period has begun. done in asm just 
								//because:)
		ldrh	r1, [r0]
		cmp	r1, #160
		bne 	scanline_wait
	}*/		
	while((volatile u16)REG_VCOUNT != 160){}	
}

///Copy our sprite array to OAM
void CopyOAM(void)
{
	u16 loop;
	u16* temp;
	temp = (u16*)sprites;
	for(loop = 0; loop < 128*4; loop++)
	{
		OAM[loop] = temp[loop];
	}
}

///Test for key presses
void GetInput(void)
{
	//this function does a lot more than just get input as you can tell.  It also
	//manages sprite movement and animation and even background scrolling
	
	//this is a static variable that decrements once per frame to determine if it is
	//time for the next animation frame
	static s16 fChangeFrame = 10;
	//these are the animation frames for pacman
	const char frameMatrix[] = {1,0,3,2,5,4,7,6};
	
//if ten frames have passed then get next animation frame
	if(fChangeFrame <= 0)
	{
		pacman.curFrame = frameMatrix[pacman.curFrame];//get next frame
		sprites[pacman.oamEntry].attribute2 = pacman.frameCharName[pacman.curFrame];//set next frame
		fChangeFrame = 10;//reset frame counter
	}
	
	if(!(*KEYS & KEY_UP))
	{
 		pacman.y-=1<<8;  //x and y are fixed point to enable finer movement than 1 pixle per
							//frame.  Not realy used in this game though
		fChangeFrame --;  //decrement frame counter
		
		if(pacman.curFrame != 4 && pacman.curFrame !=5)
		{
			//if we were not allready moving up then set the animation frame
			//to up
			pacman.curFrame = 4;
			sprites[pacman.oamEntry].attribute2 = pacman.frameCharName[pacman.curFrame];
	
		}
		
		//this code scrolles the backround as long as pacman is not near the end of the screen
		if(pacman.y >= 72<<8 && pacman.y <= (LVL_1_H-88)<<8)
			bg2.y_scroll--;
	}
	if(!(*KEYS & KEY_DOWN))
	{
		//the other keys behave simalerly
 		pacman.y+=1<<8;
		fChangeFrame --;
		if(pacman.curFrame != 6 && pacman.curFrame != 7)
		{
			pacman.curFrame = 6;
			sprites[pacman.oamEntry].attribute2 = pacman.frameCharName[pacman.curFrame];
		}
		if(pacman.y >= 72<<8 && pacman.y <= (LVL_1_H-88)<<8)
			bg2.y_scroll++;
	}
	if(!(*KEYS & KEY_LEFT))
	{
 		pacman.x-=1<<8;
		fChangeFrame --;
		if(pacman.curFrame != 2 && pacman.curFrame != 3)
		{
			pacman.curFrame = 2;
			sprites[pacman.oamEntry].attribute2 = pacman.frameCharName[pacman.curFrame];
	
		}
		if(pacman.x >= 112<<8 && pacman.x <= (LVL_1_W-128)<<8)
			bg2.x_scroll--;
	}
	if(!(*KEYS & KEY_RIGHT))
	{
 		pacman.x+=1<<8;
		fChangeFrame --;
		if(pacman.curFrame != 0 && pacman.curFrame != 1)
		{
			pacman.curFrame = 0;
			sprites[pacman.oamEntry].attribute2 = pacman.frameCharName[pacman.curFrame];
		}
		if(pacman.x >= 112<<8 && pacman.x <= (LVL_1_W-128)<<8)
			bg2.x_scroll++;
	}
	if(!(*KEYS & KEY_L))
	{
 		angle--;
		if(angle<0)
			angle = 359;
	}
	if(!(*KEYS & KEY_R))
	{
 		angle++;
		if(angle > 359)
			angle = 0;
	}
	if(!(*KEYS & KEY_A))
	{
 		zoom--;
	}
	if(!(*KEYS & KEY_B))
	{
 		zoom++;
	}
	if(!(*KEYS & KEY_START))
	{
 		
	}
	if(!(*KEYS & KEY_SELECT))
	{
 	
	}
	
//these just wrap pacman to the other side if he is out of bounds
//also resets the scroll
	if(pacman.x > (LVL_1_W<<8))
	{
		pacman.x = -15<<8;
		bg2.x_scroll = 120;
	}
	if(pacman.x < (-15<<8))
	{
		pacman.x = LVL_1_W<<8;
		bg2.x_scroll = LVL_1_W - 120;
	}
	if(pacman.y > (LVL_1_H<<8))
	{
		pacman.y = -15<<8;
		bg2.y_scroll = 0;
	}
	if(pacman.y < (-15<<8))
	{
		pacman.y = LVL_1_H<<8;
		bg2.y_scroll = LVL_1_H - 80;
	}

//make sue we do not scroll the background too far
	if(bg2.x_scroll < 0 )
		bg2.x_scroll = 80;
		
	if(bg2.x_scroll > (LVL_1_W - 120) )
		bg2.x_scroll = LVL_1_W - 120;
	
	if(bg2.y_scroll > (LVL_1_H - 80) )
		bg2.y_scroll = LVL_1_H -80;
		
	if(bg2.y_scroll < 80 )
		bg2.y_scroll = 80;



}

//move the sprite
void MoveSprite(pSprite sp)
{
//this function just adjust the x and y for the background scroll and puts it
//in our OAM copy
	FIXED x,y;
	
	x = sp->x - ((bg2.x_scroll-120)<<8); //remember fixed point numbers
	y = sp->y - ((bg2.y_scroll-80)<<8);

	if(x < 0 )
		x = (512<<8) + x;
	if(y < 0)
		y = (256<<8) + y;

//check to make sure x and y are within there bounds then copy the data in
	if(y < 256<<8 && x < 512<<8 && y > 0 && x > 0)
	{
		sprites[sp->oamEntry].attribute0 &= 0xFF00;
		sprites[sp->oamEntry].attribute1 &= 0xFE00;
		sprites[sp->oamEntry].attribute0 |= (y>>8);
		sprites[sp->oamEntry].attribute1 |= (x>>8);
	}

}	
void RotateSprite(int rotDataIndex, int angle, FIXED x_scale,FIXED y_scale)
{
	
	FIXED pa,pb,pc,pd;

	pa = ((x_scale) * COS[angle])>>8;    //(do my fixed point multiplies and shift back down)
	pb = ((y_scale) * SIN[angle])>>8;
	pc = ((x_scale) * -SIN[angle])>>8;
	pd = ((y_scale) * COS[angle])>>8;


	rotData[rotDataIndex].pa = pa;  //put them in my data struct
	rotData[rotDataIndex].pb = pb;
	rotData[rotDataIndex].pc = pc;
	rotData[rotDataIndex].pd = pd;
}

void SetUpSprite(pSprite sp)
{
	sprites[sp->oamEntry].attribute0 = COLOR_256 | SQUARE| ROTATION_FLAG |(sp->y>>8);
	sprites[sp->oamEntry].attribute1 = SIZE_16 | (sp->x>>8) | ROTDATA(sp->rot_index);
	sprites[sp->oamEntry].attribute2 = sp->frameCharName[sp->curFrame];
}


void MoveEnemies(pSprite sp)
{
	//all right this function just does some very simple AI
	//it determins if the ghost are further away in the x or y direction
	//then reduces that difference

	u16 dx,dy;  //difference in x and y position between the sprites
		
	if(sp->x > pacman.x)
		dx = sp->x - pacman.x;
	else
		dx = pacman.x - sp->x;

	if(sp->y > pacman.y)
		dy = sp->y - pacman.y;
	else
		dy = pacman.y - sp->y;

	//move the sprite
	if(dx > dy)
	{
		if(sp->x > pacman.x)
			sp->x -= sp->frameCharName[0] << 1;
		else
			sp->x += sp->frameCharName[0] << 1;
	
	}
	else
	{
		if(sp->y > pacman.y)
			sp->y -= sp->frameCharName[0] << 1;
		else
			sp->y += sp->frameCharName[0] << 1;
	
	}

//you may be wondering why I use framCharName to subtract from the x and y.  
	//really it is just because I wanted them to move at different speeds
	//and those values are different for each ghost.  Cleaner ways to do things
	//I know.

}

